#!/usr/bin/env python3
import pymatgen as mg
import pymatgen.core.units
import math
import numpy as np
import io
import textwrap
from collections import namedtuple

from typing import Dict, List, Any

from sklearn import linear_model

from . import utils

from . import rep_spline
from . import rep_basis

class RepulsiveModelCollection:
    """ General class for the collection of the repulsive potential model

    """

    def __init__(self, rep_names, rep_models):
        self.rep_names = []
        self.rep_models = []
        self.resulting_repulsives = []
        self.all_names = {}
        self.shifts = {}
        self.max_cutoff = 0.0

        s_rep_names = []
        s_rep_model = []
        for name, model in zip(rep_names, rep_models):
            elems = name.split('-')
            sorted_name = '{}-{}'.format(*sorted(elems))
            s_rep_names.append(sorted_name)
            s_rep_model.append(model)

        shift = 0
        for name, model in sorted(zip(s_rep_names, s_rep_model), key=lambda x: x[0]):
            if (name not in self.all_names):
                self.all_names[name] = model
                elems = name.split('-')
                reversed_name = '{}-{}'.format(*reversed(elems))
                self.all_names[reversed_name] = model
                self.rep_names.append(name)
                self.rep_models.append(model)
                self.shifts[name] = shift
                self.shifts[reversed_name] = shift
                self.max_cutoff = max(self.max_cutoff, model.cutoff)
                shift += model.num_variables
        
        self.total_num_variables = sum([x.num_variables for x  in self.rep_models])

    def description(self):
        fout = io.StringIO()
        print('Potential type: {}'.format(self.rep_models[0].description()), file=fout)
        print('Number of repulsive potentials: {}'.format(len(self.rep_models)), file=fout)
        print('Number of variables: {}'.format(self.get_total_num_variables()), file=fout)
        
        print('Repulsive potenial description:', file=fout)
        for rep_name, rep_model in zip(self.rep_names, self.rep_models):
            print('  {}'.format(rep_name), file=fout)
            print(textwrap.indent(str(rep_model), "    "), file=fout)

        print("", file=fout)
        return fout.getvalue()

    def get_shift(self, name):       
        return self.shifts.get(name, -1)

    def get_num_variables(self, name):
        return self.all_names[name].num_variables        
    
    def get_total_num_variables(self):
        return self.total_num_variables

    def get_row(self, name, r, derv=0):
        return self.all_names[name].get_row(r, derv)
    

class EquationDescription:
    def __init__(self, equ_type: str, equation_item, equ_block, ref_block) -> None:
        self.equ_type = equ_type
        self.equation_item = equation_item
        self.equ_block = equ_block
        self.ref_block = ref_block

class EnergyEquationDescription(EquationDescription):
    def __init__(self, equation_item, equ_block, ref_block, missing_atom_counts):
        super().__init__("energy", equation_item, equ_block, ref_block)
        self.missing_atom_counts = missing_atom_counts
        

class RepulsivePotentialEquationBuilder():
    def __init__(self, evaluated_erepfit_input):

        self.evaluated_erepfit_input = evaluated_erepfit_input

        #building a haspmap for easy reference.
        self.uuidmap = {}
        for system in self.evaluated_erepfit_input.systems:
            self.uuidmap[system.uuid] = system

        #equations for each of systems        
        self.equation_descriptions = []

        #load external repulsive potentials
        self.external_repulsive_potentials = { k: rep_spline.RepulsivePotenial.from_file(v) for k, v in self.evaluated_erepfit_input.external_repulsive_potentials.items()}
    
    def calc_dist_matrix(self, mole, cutoff):
        """ Get the distance and displacement matrix in bohr for all pairs
        """ 
        res = {}
        cutoff_ang = cutoff/mg.core.units.ang_to_bohr
        if (isinstance(mole, mg.Structure)):
            for ind, site in enumerate(mole.sites):
                neighbors = mole.get_neighbors(site, cutoff_ang, include_index=True)
                for n_site, n_r, n_index in neighbors:
                    r_in_bohr = n_r*mg.core.units.ang_to_bohr
                    vec = ((site.coords - n_site.coords)/n_r)
                    if ( (ind, n_index) in res):
                        res[(ind, n_index)].append( (r_in_bohr, vec) )
                    else:
                        res[(ind, n_index)] = [ (r_in_bohr, vec) ]
        else:
            for ind, site in enumerate(mole.sites):
                for ind2, site2 in enumerate(mole.sites):
                    if (ind == ind2):
                        continue
                    vec = site.coords - site2.coords
                    n_r = math.sqrt(sum(vec*vec))
                    r_in_bohr = n_r*mg.core.units.ang_to_bohr
                    p_vec = vec/n_r
                    if (r_in_bohr > cutoff):
                        continue
                    if ( (ind, ind2) in res):
                        res[(ind, ind2)].append( (r_in_bohr, p_vec) )
                    else:
                        res[(ind, ind2)] = [ (r_in_bohr, p_vec) ]
        return res

    def get_dist_matrix(self, system, max_cutoff):
        if system.uuid in self.dist_matrix_uuids:
            dist_matrix = self.dist_matrix_uuids[system.uuid]
            return dist_matrix
        else:
            dist_matrix = self.calc_dist_matrix(system.geometry.get_coordinate_struct(), max_cutoff)
            self.dist_matrix_uuids[system.uuid] = dist_matrix
            return dist_matrix

    def make_energy_block(self, system, dist_matrix, modelcollection, external_reps):
        num_rows = 1
        equ_block = np.zeros( (num_rows, modelcollection.get_total_num_variables()) )
        ref_block = np.zeros( num_rows )
        
        for k, dis in dist_matrix.items():
            ind1, ind2 = k
            rep_name = '{}-{}'.format(system.geometry.coordinates[ind1][0], system.geometry.coordinates[ind2][0])
            shift = modelcollection.get_shift(rep_name)
            if (shift > -1):
                for r, _ in dis:
                    row = modelcollection.get_row(rep_name, r, 0)*0.5
                    lens = len(row)
                    equ_block[0, shift:shift+lens] += row
            elif (rep_name in external_reps):
                for r, _ in dis:
                    ene = external_reps[rep_name].eval(r)
                # print('Add external: {} {} {}'.format(rep_name, r_in_bohr, ene))
                    ref_block -= ene
            else:
                raise RuntimeError('Rep not found: {}'.format(rep_name))
        return equ_block, ref_block
    
    def make_force_block(self, system, dist_matrix, modelcollection, external_reps, construct_indexes):
        
        num_rows = 3*len(construct_indexes)
        if num_rows == 0:
            return [], []

        equ_block = np.zeros( (num_rows, modelcollection.get_total_num_variables()))
        ref_block = np.zeros( num_rows )
        
        construct_indexes_index = list(construct_indexes)
        for k, dis in dist_matrix.items():
            ind1, ind2 = k
            if (ind1 not in construct_indexes):
                continue
            ind = construct_indexes_index.index(ind1)
            rep_name = '{}-{}'.format(system.geometry.coordinates[ind1][0], system.geometry.coordinates[ind2][0])
            shift = modelcollection.get_shift(rep_name)
            if (shift > -1):
                for r, vec in dis:
                    row = modelcollection.get_row(rep_name, r, 1)
                    lens = len(row)
                    for xyz in range(3):
                        equ_block[ ind*3+xyz , shift:shift+lens ] += (-row*vec[xyz])
                    
            elif (rep_name in external_reps):
                for r, vec in dis:
                    values = external_reps[rep_name].eval(r, 1)
                    for xyz in range(3):
                        ref_block[ind*3+xyz] -= -values*ref_block[xyz]
            else:
                raise RuntimeError('Rep not found: {}'.format(rep_name))
            
        return equ_block, ref_block


    def build_linear_equations(self, modelcollection: RepulsiveModelCollection):
        
        self.modelcollection = modelcollection
        self.fitted_atom_energy: Dict[str, float] = {}

        # get the maximal cutoff of all repulsive potentials
        max_cutoff:float = max([x.cutoff for x in self.modelcollection.rep_models])
        if (len(self.external_repulsive_potentials) > 0):
            max_cutoff = max(max_cutoff, max([v.cutoff for k, v in self.external_repulsive_potentials.items()]))

        self.dist_matrix_uuids :Dict[str, Dict[tuple, list]] = {}
        
        #atomization energy equations
        self._build_energy_equations(max_cutoff)

        #force equations
        self._build_force_equations(max_cutoff)
    
        #additional equations
        self._build_additional_equations()
        
        #reaction equations
        self._build_reaction_equations(max_cutoff)

    def _build_reaction_equations(self, max_cutoff: float) -> None:
        for reaction_item in self.evaluated_erepfit_input.equations['reaction']:
            react_energy = utils.get_energy_in_au(reaction_item.energy, reaction_item.unit)

            if (reaction_item.weight < 1.0e-4):
                continue

            final_equ_block = np.zeros( (1, self.modelcollection.get_total_num_variables()) )
            final_ref_block = np.zeros( 1 )

            elec_reaction_energy = 0.0
            for item in reaction_item.products:
                system = self.uuidmap[item.uuid]
                dist_matrix = self.get_dist_matrix(system, max_cutoff)
                
                equ_block, ref_block = self.make_energy_block(system, dist_matrix, self.modelcollection, self.external_repulsive_potentials)
                final_equ_block += equ_block*item.coefficient
                final_ref_block += ref_block*item.coefficient
                elec_reaction_energy += system.elec_data['elec_energy']*item.coefficient

            for item in reaction_item.reactants:
                system = self.uuidmap[item.uuid]
                dist_matrix = self.get_dist_matrix(system, max_cutoff)

                equ_block, ref_block = self.make_energy_block(system, dist_matrix, self.modelcollection, self.external_repulsive_potentials)
                final_equ_block -= equ_block*item.coefficient
                final_ref_block -= ref_block*item.coefficient
                elec_reaction_energy -= system.elec_data['elec_energy']*item.coefficient


            final_ref_block += react_energy - elec_reaction_energy

            desc = EquationDescription('reaction', reaction_item, final_equ_block*reaction_item.weight, final_ref_block*reaction_item.weight)
            self.equation_descriptions.append(desc)

    def _build_additional_equations(self) -> None:
        for rep_name, items in sorted(self.evaluated_erepfit_input.equations['additional'].items(), key=lambda x:x[0]):
            shift = self.modelcollection.get_shift(rep_name)
            if (shift > -1):
                equ_block = np.zeros( (len(items), self.modelcollection.get_total_num_variables()) )
                ref_block = np.zeros( len(items) )
                for ind, item in enumerate(items):
                    r = utils.get_length_in_au(item.distance, item.unit)
                    row = self.modelcollection.get_row(rep_name, r, item.derivative)
                    lens = len(row)
                    equ_block[ind, shift:shift+lens] += row*item.weight
                    ref_block[ind] = item.value*item.weight

                desc = EquationDescription('additional', item, equ_block, ref_block)
                desc.rep_name = rep_name
                self.equation_descriptions.append(desc)
            else:
                continue

    def _build_force_equations(self, max_cutoff: float) -> None:
        for item in self.evaluated_erepfit_input.equations['force']:
            if (item.weight < 1.0e-4):
                continue
            # Get the bond distances.
            system = self.uuidmap[item.uuid]
            dist_matrix = self.get_dist_matrix(system, max_cutoff)

            fitting_indexes = list(range(len(system.geometry.coordinates)))
            if hasattr(item, 'reference_force'):
                if hasattr(item, 'fitting_indexes'):
                    fitting_indexes = item.fitting_indexes
                else:
                    item['fitting_indexes'] = fitting_indexes

            equ_block, ref_block = self.make_force_block(system, dist_matrix, self.modelcollection, self.external_repulsive_potentials, fitting_indexes)


            if (len(ref_block) == 0):
                continue

            total_forces = np.zeros( ref_block.shape )
            if hasattr(item, 'reference_force'):
                total_forces = np.array([item.reference_force[i] for i in fitting_indexes]).flatten()

            ref_e_forces = np.array([ system.elec_data['elec_forces'][i] for i in fitting_indexes ]).flatten()
            # print(ref_block.shape, ref_e_forces.shape, total_forces.shape)
            ref_block =  (total_forces -  ref_e_forces) - ref_block

            desc = EquationDescription('forces', item, equ_block*item.weight, ref_block*item.weight)
            self.equation_descriptions.append(desc)

    def _build_energy_equations(self, max_cutoff: float) -> None:
        for item in self.evaluated_erepfit_input.equations['energy']:
            if (item.weight < 1.0e-4):
                continue
            # Get the bond distances.
            system = self.uuidmap[item.uuid]
            dist_matrix = self.get_dist_matrix(system, max_cutoff)

            missing_atom_counts = {}
            e_atom = 0.0
            for line in system.geometry.coordinates:
                symbol = line[0]
                if symbol not in self.evaluated_erepfit_input.atomic_energy:
                    value = missing_atom_counts.get(symbol, 0)
                    missing_atom_counts[symbol] = value+1
                else:
                    e_atom += self.evaluated_erepfit_input.atomic_energy[symbol]

            e_ea_dft = utils.get_energy_in_au(item.energy, item.unit)
            e_rep_dftb = e_atom - system.elec_data['elec_energy'] - e_ea_dft

            equ_block, ref_block = self.make_energy_block(system, dist_matrix, self.modelcollection, self.external_repulsive_potentials)
            ref_block = -ref_block + e_rep_dftb

            desc = EnergyEquationDescription(item, equ_block*item.weight, ref_block*item.weight, missing_atom_counts)
            self.equation_descriptions.append(desc)


    

    def _extend_atomic_energy_equations(self, all_missing_atoms, matrix) -> None:
        if (len(all_missing_atoms) == 0):
            return matrix
        atom_columns = len(all_missing_atoms)
        new_matrix = np.zeros( (matrix.shape[0], matrix.shape[1]+atom_columns) ) 
        new_matrix[:, :matrix.shape[1]] = matrix

        ind = 0
        for desc in self.equation_descriptions:
            if isinstance(desc, EnergyEquationDescription):
                for i in range(atom_columns):
                    new_matrix[ind, matrix.shape[1]+i] = -(desc.missing_atom_counts.get(all_missing_atoms[i], 0))*desc.equation_item.weight            
            ind += desc.equ_block.shape[0]
        return new_matrix
        

    
    def get_residuals(self) -> List[Any]:
        fout = io.StringIO()
        return sum([sum(res.residual*res.residual) for res in self.print_residuals(fout)])

    def print_residuals(self, output_stream) -> List[Any]:
        Residual = namedtuple('Residual', ['description', 'residual'])
        residuals = []
        print("", file=output_stream)

        if (len(self.fitted_atom_energy) > 0):
            print('Fitted Atom Energies:', file=output_stream)
            for k, v in sorted(self.fitted_atom_energy.items()):
                print(' {:<3s}: {:20.12f} Hartree'.format(k,v), file=output_stream)
        print("", file=output_stream)

        print("Energy Equation Residuals", file=output_stream)
        vunknowns = np.array(self.vunknowns)
        for desc in [x for x in self.equation_descriptions if x.equ_type == 'energy']:
            item = desc.equation_item
            atom_energy = sum([self.fitted_atom_energy[k]*v for k, v in desc.missing_atom_counts.items()]) * item.weight
            res = (np.dot(desc.equ_block, vunknowns) - atom_energy - desc.ref_block) / item.weight
            residuals.append(Residual(desc, res))
            print('{:10.4f} {:9s} {}'.format(utils.get_energy_from_au(res[0], item.unit), item.unit, item.name), file=output_stream)
        
        print("", file=output_stream)
        print("Force Equation Residuals (Ha/bohr)", file=output_stream)
        for desc in [x for x in self.equation_descriptions if x.equ_type == 'forces']:
            item = desc.equation_item
            res = ((np.dot(desc.equ_block, vunknowns) - desc.ref_block) / item.weight)
            print(item.name, item.weight, file=output_stream)
            system = self.uuidmap[item.uuid]
            residuals.append(Residual(desc, res))
            if hasattr(item, 'fitting_indexes'):
                for i in range(0, int(len(res)/3)):
                    print(' {:3s} {:20.12f} {:20.12f} {:20.12f}'.format(system.geometry.coordinates[item.fitting_indexes[i]][0], 
                        res[i*3], res[i*3+1], res[i*3+2]), file=output_stream)
            else:
                for i in range(0, int(len(res)/3)):
                    print(' {:3s} {:20.12f} {:20.12f} {:20.12f}'.format(system.geometry.coordinates[i][0], 
                        res[i*3], res[i*3+1], res[i*3+2]), file=output_stream)
            print("", file=output_stream)

        print("", file=output_stream)
        print("Addition Equation Residuals (Ha)", file=output_stream)
        for desc in [x for x in self.equation_descriptions if x.equ_type == 'additional']:
            item = desc.equation_item
            rep_name = desc.rep_name
            res = np.dot(desc.equ_block, vunknowns) - desc.ref_block           
            residuals.append(Residual(desc, res))
            for i in range(len(res)):
                print('  {:<6s} {:20.12f}'.format(rep_name, res[i]), file=output_stream)

        print("", file=output_stream)
        print("Reaction Equation Residuals (Ha)", file=output_stream)
        for desc in [x for x in self.equation_descriptions if x.equ_type == 'reaction']:
            item = desc.equation_item
            res = (np.dot(desc.equ_block, vunknowns) - desc.ref_block ) / item.weight
            residuals.append(Residual(desc, res))
            for i in range(len(res)):
                print('{:10.4f} {:9s} {}'.format(utils.get_energy_from_au(res[i], item.unit), item.unit, item.name), file=output_stream)
        print("", file=output_stream)
        return residuals



    def solve(self, modelcollection, solver):
        self.vunknowns = solver.solve(modelcollection, self)
        



import pytest
from pyerepfit.utils import *
from unittest.mock import patch

@pytest.mark.parametrize("value, unit, expected", [
    (627.50947415, 'kcal/mol', 1.0),
    (1.0, 'h', 1.0),
    (27.211386, 'ev', 1.0),
    (pytest.param(2.0, 'Ry', 1.0, marks=[pytest.mark.xfail(raises=ValueError)]))
])
def test_get_energy_in_au(value, unit, expected):
    assert get_energy_in_au(value, unit) == pytest.approx(expected)

@pytest.mark.parametrize("value, unit, expected", [
    (1.0, 'bohr', 1.0),
    (0.529177, 'aa', 1.0),
    (pytest.param(2.0, 'cm', 1.0, marks=[pytest.mark.xfail(raises=ValueError)]))
])
def test_get_length_in_au(value, unit, expected):
    assert get_length_in_au(value, unit) == pytest.approx(expected)

@pytest.mark.parametrize("expected, unit, value", [
    (627.50947415, 'kcal/mol', 1.0),
    (1.0, 'h', 1.0),
    (27.211386, 'ev', 1.0),
    (pytest.param(2.0, 'Ry', 1.0, marks=[pytest.mark.xfail(raises=ValueError)]))
])
def test_get_energy_from_au(value, unit, expected):
    assert get_energy_from_au(value, unit) == pytest.approx(expected)

def myexists(path):
    return True

def myexists2(path):
    if (path.startswith('/def')):
        return True
    return False

def myexists3(path):
    return False


def test_which(monkeypatch):
    with monkeypatch.context() as m:
        m.setenv('PATH', '/abc:/def')
        m.setattr(os.path, 'exists', myexists)
        assert which('ala') == '/abc/ala'

def test_which2(monkeypatch):
    with monkeypatch.context() as m:
        m.setenv('PATH', '/abc:/def')
        m.setattr(os.path, 'exists', myexists2)
        assert which('ala') == '/def/ala'

def test_which3(monkeypatch):
    with monkeypatch.context() as m:
        m.setattr(os.path, 'exists', myexists3)
        assert which('ala') == 'ala'
    



#!/usr/bin/env python3

import argparse
import os
import sys
import json
import jsonschema
import yaml
import jsonpickle
import numpy as np

from . import evaluator
from .inputdata import ErepfitInput
from .rep_basis import PolynomialRepulsive, BasisRepulsiveModel
from .rep_spline import RepulsivePotenial, Spline4RepulsiveModel
from . import rep_model
from . import utils
from . import equation_solver
import os

from pyerepfit import __version__

class Erepfit:

    def __init__(self, output_stream=sys.stdout):
        self.output_stream = output_stream

    def print_header(self):
        print(
"""
===============================================================================

        PyErepfit v{}
        Powered by Chien-Pin Chou

===============================================================================
""".format(__version__), file=self.output_stream
        )

    def load_input(self, options):
        parser = argparse.ArgumentParser()
        parser.add_argument("-y", dest='yaml', help='Parse input as YAML format', action='store_true')
        parser.add_argument("-e", dest='evaluate_only', help='Evaluate the DFTB only', action='store_true')
        parser.add_argument("-f", dest='fitting_only', help='Fitting the repulsive only', action='store_true')
        self.hasPrint = False
        try:
            import matplotlib.pyplot
            parser.add_argument("-p", dest='print', help='Print the repulsive to eps', action='store_true')
            self.hasPrint = True
        except:
            pass
        parser.add_argument("-o", dest='output_file', help='Output file', type=str)
        parser.add_argument("input_file_path", help="Input file path")

        self.opt = parser.parse_args(options)

        if (self.opt.output_file):
            self.output_stream = open(self.opt.output_file, 'w')
    
        print('Reading input schema.', file=self.output_stream)
        with open(os.path.join(os.path.dirname(__file__), 'erepfit_input_schema.json'), 'r') as f:
            input_schema = json.load(f)

        if (self.opt.yaml):
            print('Reading input file as YAML', file=self.output_stream)
            with open(self.opt.input_file_path, 'r') as f:
                input_file = yaml.load(f)
        else:
            print('Reading input file as JSON', file=self.output_stream)
            with open(self.opt.input_file_path, 'r') as f:
                input_file = json.load(f)

        print('Validate input file.', file=self.output_stream)
        jsonschema.validate(input_file, input_schema)

        self.input = ErepfitInput(input_file)

    def initialize_solver(self):
        if (self.input.options.get("potential_type") == "polynomial"):
            self.solver = equation_solver.GeneralSolver()
        else:
            self.solver = equation_solver.ContinuitySolver(3)

    def initialize_skfile_path(self):
        self.input.electronic_slater_koster_files.adjust_path(os.path.abspath(os.curdir))

    def initialize_model(self):       
        rep_models = []
        rep_names = []
        poten_type = self.input.options.get("potential_type")
        
        for k, v in self.input.potential_grids.items():
            if (k in self.input.external_repulsive_potentials):
                print('Repulsive {} found in external repulsuive potentials, will not be fitted.'.format(k), file=self.output_stream)
                continue
            if (poten_type == "polynomial"):
                cutoff = v['knots'][-1]
                model = BasisRepulsiveModel(cutoff, minimal_order=3, order=9, start=v['knots'][0])
            else: 
                model = Spline4RepulsiveModel(v['knots'])
            rep_names.append(k)
            rep_models.append(model)

        self.collection = rep_model.RepulsiveModelCollection(rep_names, rep_models)      

        print("", file=self.output_stream)
        print(self.collection.description(), file=self.output_stream)

    def evaluate_elec_data(self):
        print('Testing DFTB implementation for parameterization...', file=self.output_stream)
        dftb_path = self.input.options['toolchain']['path']
        if not os.path.isabs(dftb_path):
            dftb_path = os.path.expanduser(dftb_path)
        dftb_path = os.path.expandvars(dftb_path)
        if not os.path.isabs(dftb_path):
            dftb_path = utils.which(dftb_path)

        succ = evaluator.test_dftb_binary(dftb_path)
        if (succ):
            print('  {} => working'.format(dftb_path), file=self.output_stream)
        else:
            print('  {} => not working'.format(dftb_path), file=self.output_stream)
            sys.exit(1)
        
        evaluator.evaluate(self.input)
        print('\nEvaluate DFTB...done', file=self.output_stream)

        jsonpickle.set_encoder_options('simplejson', indent=4)
        jsonpickle.set_encoder_options('json', indent=4)
        
        filename = 'evaluated_{}.json'.format(os.path.basename(self.opt.input_file_path))
        with open(filename, 'w') as f:
            js = jsonpickle.encode(self.input, unpicklable=False)
            print(js, file=f)
            print('Evaluated input file has been saved as {}'.format(filename))

    def build(self):
        self.builder = rep_model.RepulsivePotentialEquationBuilder(self.input)
        self.builder.build_linear_equations(self.collection)

    def write_bond_statistic_graphs(self, filename='bond_length.pdf'):
        from matplotlib.backends.backend_pdf import PdfPages
        import matplotlib.pyplot as plt
        import datetime
        pair_bond_data = {}
        for uuid, dist_matrix in self.builder.dist_matrix_uuids.items():
            system = self.builder.uuidmap[uuid]
            for index_pair, dist_vector in dist_matrix.items():
                elem1 = system.geometry.coordinates[index_pair[0]][0]
                elem2 = system.geometry.coordinates[index_pair[1]][0]
                if elem1 > elem2 :
                    continue
                poten_name = '{}-{}'.format(elem1, elem2)
                if poten_name not in self.final_repulsive_potentials:
                    continue
                distances = [x[0] for x in dist_vector if x[0] <= self.final_repulsive_potentials[poten_name].knots[-1]]
                if (elem1, elem2) not in pair_bond_data:
                    pair_bond_data[(elem1, elem2)] = np.array([])
                    
                pair_bond_data[(elem1, elem2)] = np.append(pair_bond_data[(elem1, elem2)], distances)
        with PdfPages(filename) as pdf:
            for k, v in pair_bond_data.items():
                plt.figure(figsize=(5, 5))
                plt.hist(v, bins=50)
                plt.xlabel('Distance [bohr]')
                plt.title('{}-{}'.format(*k))
                pdf.savefig()  
                plt.close()

            d = pdf.infodict()
            d['Title'] = 'Bond Length'
            d['CreationDate'] = datetime.datetime.today()

    def solve(self):
        self.builder.solve(self.collection, self.solver)
    
    def print_residuals(self):
        self.builder.print_residuals(self.output_stream)

        self.final_repulsive_potentials = {}
        for ind in range(len(self.collection.rep_names)):
            rep_name = self.collection.rep_names[ind]
            spl_rep = self.collection.resulting_repulsives[ind]
            reversed_rep_name = '{}-{}'.format(*reversed(rep_name.split('-')))
            self.final_repulsive_potentials[rep_name] = spl_rep
            self.final_repulsive_potentials[reversed_rep_name] = spl_rep

    def write(self):
        print('Resulting Repulsive Potentials:', file=self.output_stream)
        output_prefix = self.input.options.get("output_prefix", "./")
        os.makedirs(output_prefix, exist_ok=True)
        
        for ind in range(len(self.collection.rep_names)):
            rep_name = self.collection.rep_names[ind]
            spl_rep = self.collection.resulting_repulsives[ind]
            reversed_rep_name = '{}-{}'.format(*reversed(rep_name.split('-')))
        
            with open(os.path.join(output_prefix, '{}.skf'.format(rep_name)), 'w') as fout:
                print(spl_rep, file=fout)
            if (reversed_rep_name != rep_name):
                with open(os.path.join(output_prefix, '{}.skf'.format(reversed_rep_name)), 'w') as fout:
                    print(spl_rep, file=fout)

            print(rep_name, file=self.output_stream)
            print(spl_rep, file=self.output_stream)

            if (self.hasPrint and self.opt.print):
                import matplotlib.pyplot as plt
                plt.figure()
                output_fig = os.path.join(output_prefix, '{}.eps'.format(rep_name))

                plt.title(rep_name)
                # plt.ylim(-2, 2)

                styles = ['c--', 'r--', 'g--']
                labels = ['Rep. Potential', '1st derivative', '2nd derivative']
                for i in range(3):
                    xs = np.linspace(spl_rep.knots[0]-0.2, spl_rep.knots[-1], num=100)
                    ys = []
                    for r in xs:
                        v = spl_rep.eval(r, i)
                        ys.append(v)
                    if (i==0):
                        maxv = max(ys)
                        minv = min(ys)
                        plt.ylim(min(minv, max(-0.5, -maxv)), maxv)
                    plt.plot(xs, ys, styles[i], linewidth=1, label=labels[i])

                plt.legend(loc='upper right')
                plt.xlabel('Diatomic Distance [a.u.]')
                plt.ylabel('Energy [a.u.]')
                plt.savefig(output_fig, format='eps')


    def run(self, options):       
        self.print_header()
        self.load_input(options)
        
        # if only evaluate electronic reference values
        if (self.opt.evaluate_only or not self.opt.fitting_only):
            self.initialize_skfile_path()
            self.evaluate_elec_data()

        if (self.opt.fitting_only or not self.opt.evaluate_only):
            self.initialize_model()
            self.initialize_solver()
            self.build()
            self.solve()
            self.print_residuals()
            self.write()
            self.write_bond_statistic_graphs()

        print('End of pyerepfit Program', file=self.output_stream)        
        

if __name__ == '__main__':
    erepfit = Erepfit()
    erepfit.run(sys.argv[1:])
    






#!/usr/bin/env python3
import numpy as np
from .rep_spline import RepulsivePotenial
from scipy.interpolate import CubicSpline

class BasisRepulsiveModel:
    def __init__(self, cutoff, minimal_order = 3, order = 9, start = 1.0):
        self.cutoff = cutoff
        self.minimal_order = minimal_order
        self.order = order
        self.num_variables = self.order - self.minimal_order + 1
        self.start = start
    def get_row(self, r, derv=0):
        res = np.zeros(self.num_variables)
        for i in range(self.num_variables):
            if (derv == 0):
                res[i] =  (self.cutoff-r)**(i+self.minimal_order)
            elif (derv == 1):
                res[i] =  -1.0*(i+self.minimal_order)*(self.cutoff-r)**(i+self.minimal_order-1)
            elif (derv == 2):
                res[i] =  (i+self.minimal_order)*(i+self.minimal_order-1)*(self.cutoff-r)**(i+self.minimal_order-2)
            else:
                raise ValueError('Not valid derivative')
        return res
    def description(self):
        return "Polynomial-type repulsive potential (R-R_cutoff)^n"
    
    def __str__(self):
        return "Minimal and maximal orders: {}, {}.\nCutoff: {}".format(self.minimal_order, self.order, self.cutoff)

class PolynomialRepulsive:
    def __init__(self, name, cutoff, coeffs, minimal_order=3, start=1.0):
        self.name = name
        self.cutoff = cutoff
        self.coeffs = coeffs
        self.minimal_order = minimal_order
        self.start = start
    
    def eval(self, distance: float, derivs=0):
        if (distance >= self.cutoff):
            return 0.0
        func = 0.0
        if (derivs == 0):
            for i, c in enumerate(self.coeffs):
                func += c*(self.cutoff-distance)**(self.minimal_order+i) 
        elif (derivs==1):
            for i, c in enumerate(self.coeffs):
                func += -1.0*(self.minimal_order+i)*c*(self.cutoff-distance)**(self.minimal_order+i-1) 
        elif (derivs==2):
            for i, c in enumerate(self.coeffs):
                func += (self.minimal_order+i)*(self.minimal_order+i-1)*c*(self.cutoff-distance)**(self.minimal_order+i-2) 
        else:
            raise ValueError('Derivative out of range')
        return func

    def to_spline4(self, interval=0.05):
        num = (self.cutoff - self.start)/interval + 1
        xs = np.linspace(self.start, self.cutoff, num)
        ys = [self.eval(r, 1) for r in xs]
        d1 = self.eval(xs[0], 2)
        rep = CubicSpline(xs, ys, bc_type=((1, d1), (1, 0.0)))
        spline_coeffs = []
        for i in range(len(rep.c[0])):
            l_coeffs = [ self.eval(xs[i], 0), rep.c[3][i], rep.c[2][i]/2.0, rep.c[1][i]/3.0, rep.c[0][i]/4.0]
            spline_coeffs.append(l_coeffs)
        spl_rep = RepulsivePotenial(self.name, xs, spline_coeffs)
        return spl_rep



#!/usr/bin/env python3

import sys
import math
import io
import decimal
import os
import numpy as np
import scipy.interpolate
from math import isclose

from typing import List, Union

class Spline5:
    """Fifth-order spline function"""
    
    def __init__(self, r0: float, r1: float, a0: float, a1: float, a2: float, a3: float, a4: float, a5: float) -> None:
        self.r0 = r0
        self.r1 = r1
        self.a0 = a0
        self.a1 = a1
        self.a2 = a2
        self.a3 = a3
        self.a4 = a4
        self.a5 = a5

    def eval(self, r:float, der:int=0) -> float:
        if (r<self.r0 or r>self.r1):
            raise ValueError('r: {}. Out of range {}-{}.'.format(r, self.r0, self.r1))
        if (der <0 or der>3):
            raise ValueError('der: {}. Out of range 0-3.'.format(der))

        dx = r-self.r0
        if der == 0:
            return self.a0+self.a1*dx+self.a2*dx*dx+self.a3*dx**3+self.a4*dx**4+self.a5*dx**5
        elif der == 1:
            return self.a1+2.0*self.a2*dx+3.0*self.a3*dx*dx+4.0*self.a4*dx**3+5.0*self.a5*dx**4
        elif der == 2:
            return 2.0*self.a2+3.0*2.0*self.a3*dx+4.0*3.0*self.a4*dx*dx+5.0*4.0*self.a5*dx**3
        elif der == 3:
            return 3.0*2.0*self.a3+4.0*3.0*2.0*self.a4*dx+5.0*4.0*3.0*self.a5*dx*dx
        else:
            raise ValueError("Invalid input")
    
    def __str__(self) -> str:
        return '{:<12.6f} {:<12.6f} {:20.12E} {:20.12E} {:20.12E} {:20.12E} {:20.12E} {:20.12E}'.format(
            self.r0, self.r1,
            self.a0, self.a1, self.a2, self.a3, self.a4, self.a5
        )

class Spline4:
    """Fourth-order spline function"""
    
    def __init__(self, r0: float, r1: float, a0: float, a1: float, a2: float, a3: float, a4: float) -> None:
        self.r0 = r0
        self.r1 = r1
        self.a0 = a0
        self.a1 = a1
        self.a2 = a2
        self.a3 = a3
        self.a4 = a4

    def eval(self, r: float, der:int=0) -> float:
        if (r<self.r0 or r>self.r1):
            raise ValueError('r: {}. Out of range {}-{}.'.format(r, self.r0, self.r1))
        if (der <0 or der>3):
            raise ValueError('der: {}. Out of range 0-3.'.format(der))

        dx = r-self.r0
        if der == 0:
            return self.a0+self.a1*dx+self.a2*dx*dx+self.a3*dx*dx*dx+self.a4*dx*dx*dx*dx
        elif der == 1:
            return self.a1+2.0*self.a2*dx+3.0*self.a3*dx*dx+4.0*self.a4*dx*dx*dx
        elif der == 2:
            return 2.0*self.a2+3.0*2.0*self.a3*dx+4.0*3.0*self.a4*dx*dx
        elif der == 3:
            return 3.0*2.0*self.a3+4.0*3.0*2.0*self.a4*dx
        else:
            raise ValueError("Invalid input")
    def __str__(self) -> str:
        return '{:<12.6f} {:<12.6f} {:20.12E} {:20.12E} {:20.12E} {:20.12E} {:20.12E}'.format(
            self.r0, self.r1,
            self.a0, self.a1, self.a2, self.a3, self.a4
        )

class Spline3:
    """Cubic spline function"""
    
    def __init__(self, r0: float, r1: float, a0: float, a1: float, a2: float, a3: float) -> None:
        self.r0 = r0
        self.r1 = r1
        self.a0 = a0
        self.a1 = a1
        self.a2 = a2
        self.a3 = a3

    def eval(self, r: float, der:int=0) -> float:
        if (r<self.r0 or r>self.r1):
            raise ValueError('r: {}. Out of range {}-{}.'.format(r, self.r0, self.r1))
        if (der <0 or der>2):
            raise ValueError('der: {}. Out of range 0-2.'.format(der))

        dx = r-self.r0
        if der == 0:
            return self.a0+self.a1*dx+self.a2*dx*dx+self.a3*dx*dx*dx
        elif der == 1:
            return self.a1+2.0*self.a2*dx+3.0*self.a3*dx*dx
        elif der == 2:
            return 2.0*self.a2+3.0*2.0*self.a3*dx
        else:
            raise ValueError("Invalid input")

    def __str__(self) -> str:
        return '{:<12.6f} {:<12.6f} {:20.12E} {:20.12E} {:20.12E} {:20.12E}'.format(
            self.r0, self.r1,
            self.a0, self.a1, self.a2, self.a3
        )


class RepulsivePotenial:
    """Repulsive potential in DFTB"""

    def eval(self, r:float, der:int=0) -> float:
        if (r < self.splines[0].r0):
            if (der == 0):
                return math.exp(-self.expA*r+self.expB) + self.expC
            elif (der == 1):
                return -self.expA*math.exp(-self.expA*r+self.expB) 
            elif (der == 2):
                return self.expA*self.expA*math.exp(-self.expA*r+self.expB) 
            elif (der == 3):
                return (-self.expA**3)*math.exp(-self.expA*r+self.expB) 
        
        # ! todo: write the test and test this
        # if (r >= self.knots[-1]):
        #     return 0.0

        # index = np.searchsorted(self.knots[:-1], r, side='right') - 1 
        # self.splines[index].eval(r, self.derv)

        for spline in self.splines:
            if (r >= spline.r0 and r < spline.r1):
                return spline.eval(r, der)
        return 0.0


    def fix_exponential(self) -> None:
        found = False
        r = self.knots[0]
        max_r = 2.0
        last_good = None
        while(r < max_r):
            new_expA = -self.eval(r,2)/self.eval(r,1) 
            try:
                new_expB = math.log(-self.eval(r, 1)/new_expA) + new_expA*r
                new_expC = self.eval(r, 0) - math.exp(-new_expA*r+new_expB)
                last_good = (new_expA, new_expB, new_expC)
                if (new_expA < 1.0):
                    r += 0.01
                    continue
                found = True
                break
            except:
                pass
            r += 0.01
        
        if (found == False and last_good is None):
            return
        
        new_expA, new_expB, new_expC = last_good
        new_splines = []
        found_spline = False
        self.expA = new_expA
        self.expB = new_expB
        self.expC = new_expC
        for spline in self.splines:
            if (r >=spline.r0 and r< spline.r1):
                b0 = spline.eval(r,0)
                b1 = spline.eval(r,1)
                b2 = spline.eval(r,2)/2.0
                b4 = spline.a4
                b3 = 4*b4*(r-spline.r0)+spline.a3
                new_splines.append(Spline4(r, spline.r1, b0, b1, b2, b3, b4))
                found_spline = True
            else:
                if (found_spline == False):
                    continue
                else:
                    new_splines.append(spline)
        self.splines = new_splines
        self.nsplines = len(self.splines)

    def __init__(self, name: str, knots: List[float], coeffs: List[List[float]], calc_exp:bool=True) -> None:
        self.name = name
        self.nsplines = len(knots)-1
        if len(coeffs) != self.nsplines:
            print(len(coeffs), self.nsplines)
            raise ValueError("Size inconsistent: knots and coeffs")
        if (len(coeffs[0]) == 4 or (self.nsplines == 1 and len(coeffs[0])==6)):
            self.order = 3
        elif (len(coeffs[0]) == 5):
            self.order = 4
        else:
            raise ValueError("Size inconsistent: number of coeffs")
        
        self.splines : List[ Union[Spline3, Spline4, Spline5] ] = []
        self.knots: List[float]  = knots
        self.cutoff : float = knots[-1]

        for i in range(0, len(knots)-1):
            r0 = knots[i]
            r1 = knots[i+1]
            ind = i
            if (self.order == 3):
                if (ind == len(coeffs)-1):
                    l_coeffs = coeffs[ind]
                    if (len(l_coeffs) <6):
                        l_coeffs += [0.0]*(6-len(l_coeffs))
                        spline = Spline5(r0, r1, *l_coeffs)
                    spline = Spline5(r0, r1, *coeffs[ind])
                else:
                    spline = Spline3(r0, r1, *coeffs[ind][:4])
            else:
                spline = Spline4(r0, r1, *coeffs[ind])
            self.splines.append(spline)
        
        
        if (calc_exp):
            try:
                self.expA :float = -2.0*self.splines[0].a2 / self.splines[0].a1
                self.expB :float = math.log(-self.splines[0].a1/self.expA) + self.expA*self.splines[0].r0
                self.expC :float = self.splines[0].a0 - math.exp(-self.expA*self.splines[0].r0+self.expB)
            except:
                self.expB :float = math.nan
                self.expC :float = math.nan
                self.fix_exponential()

    def __str__(self) ->str:
        with io.StringIO() as sout:
            if (self.order == 3):
                print('Spline', file=sout)
            else:
                print('Spline4', file=sout)
            print('{} {:.6f}'.format(len(self.splines), self.splines[-1].r1), file=sout)
            print('{:.6E} {:.6E} {:.6E}'.format(self.expA, self.expB, self.expC), file=sout)
            for spline in self.splines:
                print (str(spline), file=sout)

            return sout.getvalue()
    
    @classmethod
    def from_file(cls, filename: str) -> object:
        with open(filename, 'r') as f:
            for line in f:
                if line.startswith('Spline'):
                    line = next(f)
                    arr = line.split()
                    nspline = int(arr[0])
                    cutoff = float(arr[1])
                    line = next(f)
                    arr = line.split()
                    exp_a = float(arr[0])
                    exp_b = float(arr[1])
                    exp_c = float(arr[2])
                    knots : List[float] = []
                    coeffs : List[List[float]] = []
                    for _ in range(nspline):
                        try:
                            line = next(f)
                        except StopIteration:
                            raise ValueError('Number of splines is not consistent')
                        arr = line.split()
                        x0 = float(arr[0])
                        x1 = float(arr[1])
                        cofs = list(map(lambda x: float(x), arr[2:]))
                        coeffs.append(cofs)                    
                        if (len(knots) == 0):
                            knots.append(x0)    
                        knots.append(x1)
                    
                    if not isclose(knots[-1], cutoff, abs_tol=1e-6):
                        raise ValueError('Cutoff is not consistent')

                    obj = cls(os.path.basename(filename), knots, coeffs, False)
                    obj.expA = exp_a
                    obj.expB = exp_b
                    obj.expC = exp_c
                    return obj
            raise RuntimeError('Not a valid repulsive potential.')


class Spline4Model:

    def __init__(self, x1:float, x2:float) -> None:
        self.x1 = x1
        self.x2 = x2
        self.num_variables : int = 5

    def get_row(self, r:float, derv:int=0) :
        res = np.zeros(self.num_variables)
        dx = r - self.x1
        if dx < 0.0:
            raise ValueError('r is < r0.')
        for i in range(derv, self.num_variables):
            if (derv == 0):
                res[i] = dx**i
            elif (derv == 1):
                res[i] =  (i)*dx**(i-1)
            elif (derv == 2):
                res[i] =  (i*(i-1))*dx**(i-2)
            elif (derv == 3):
                res[i] =  (i*(i-1)*(i-2))*dx**(i-3)
            else:
                raise ValueError('Not valid derivative')
        return res

class Spline4RepulsiveModel:
    def __init__(self, knots: List[float]) -> None:
        self.knots = sorted(knots)
        self.cutoff = self.knots[-1]
        self.spline_models : List[Spline4Model] = []
        for i in range(len(self.knots)-1):
            self.spline_models.append(Spline4Model(self.knots[i], self.knots[i+1]))
        self.num_variables = sum([x.num_variables for x in self.spline_models])
        
    def get_row(self, r:float, derv:int=0) :
        row = np.zeros( self.num_variables )

        if r > self.knots[-1] :
            return row
        if r < self.knots[0]:
            raise ValueError('r is < r0.')
        index = np.searchsorted(self.knots[:-1], r, side='right') - 1 
        shift = sum([x.num_variables for x in self.spline_models[:index]])
        row[shift:shift+self.spline_models[index].num_variables] = self.spline_models[index].get_row(r, derv)
        return row


    def description(self) -> str:
        return "Fourth-order spline-type repulsive potential"

    def __str__(self) -> str:
        return "Knots: {} bohr".format('{:8.4f}'*len(self.knots)).format(*self.knots)
        



#!/usr/bin/env python3

import subprocess
import tempfile
import os
import scipy.constants
import numpy
# from erepfit import *

def Bool2YesNo(value):
    if (value):
        return "Yes"
    else:
        return "No"

def evaluate_atom(dftb_path, dftb_input, skfileinfo):
    with tempfile.TemporaryDirectory() as tmp:
        with open(os.path.join(tmp, 'dftb_in.hsd'), 'w') as fout:
            #copy the template input
            with open(dftb_input, 'r') as fin:
                fout.write(fin.read())
            #add SK file info
            skfileinfo.to_hsd(fout)
            print('*ParserOptions = {', file=fout)
            print('    !IgnoreUnprocessedNodes = Yes', file=fout)
            print('}', file=fout)
        
        res = subprocess.run([dftb_path], stdout=subprocess.PIPE, cwd=tmp)
        if (res.returncode == 0):
            try:
                with open(os.path.join(tmp,'detailed.out'), 'r') as fin:
                    for line in fin:
                        if 'Total energy' in line:
                            return float(line.split()[2])
            except:
                raise RuntimeError(res.stdout.decode())
        else:
            raise RuntimeError(res.stdout.decode())            



def evaluate_system(dftb_path, system, skfileinfo, dftb_options):
    with tempfile.TemporaryDirectory() as tmp:
        with open(os.path.join(tmp, 'dftb_in.hsd'), 'w') as fout:
            #copy the template input
            with open(system.template_input, 'r') as fin:
                fout.write(fin.read())
            #add SK file info
            skfileinfo.to_hsd(fout)
            system.to_hsd(fout)

            if 'hubbard_derivatives' in dftb_options:
                print('+Hamiltonian = +DFTB {', file=fout)
                print('   *HubbardDerivs = {', file=fout)
                for k, v in dftb_options['hubbard_derivatives'].items():
                    print('        !{} = {}'.format(k,v), file=fout)
                print('    }', file=fout)
                print('}', file=fout)

            print('*Options = {', file=fout)
            print('    !WriteAutotestTag = Yes', file=fout)
            print('}', file=fout)
            print('*Analysis = {', file=fout)
            print('    !CalculateForces = Yes', file=fout)
            print('}', file=fout)
            print('*ParserOptions = {', file=fout)
            print('    !IgnoreUnprocessedNodes = Yes', file=fout)
            print('}', file=fout)
            
        
        
        res = subprocess.run([dftb_path], stdout=subprocess.PIPE, cwd=tmp)
        if (res.returncode == 0):
            try:
                with open(os.path.join(tmp,'autotest.tag'), 'r') as fin:
                    for line in fin:
                        if line.startswith('total_energy'):
                            line = next(fin)
                            total_energy = float(line)
                        elif line.startswith('repulsive_energy'):
                            line = next(fin)
                            rep_energy = float(line)
                        elif line.startswith('repulsive_forces'):
                            entry = line.split(':')[3]
                            nrow = int(entry.split(',')[1])
                            rep_forces = []
                            for i in range(nrow):
                                line = next(fin)
                                rep_forces.append(list(map(float, line.split())))
                        elif line.startswith('forces'):
                            entry = line.split(':')[3]
                            nrow = int(entry.split(',')[1])
                            forces = []
                            for i in range(nrow):
                                line = next(fin)
                                forces.append(list(map(float, line.split())))
                elec_energy = (total_energy - rep_energy)
                elec_forces = ((numpy.array(forces)-numpy.array(rep_forces))).tolist()
                # print(system.name, '{:20.12f}'.format(elec_energy))
                # for line in elec_forces:
                #     print('{:20.12f} {:20.12f} {:20.12f}'.format(*line))
                return total_energy, elec_energy, elec_forces
                            
            except:
                with open(os.path.join(tmp, 'dftb_in.hsd'), 'r') as f:
                    print(f.read())
                raise RuntimeError(res.stdout.decode())  

                
        else:
            raise RuntimeError(res.stdout.decode())  

def evaluate(erepfit_input):
    from tqdm import tqdm
    
    print('Starting evaluation...')
    print('')
    
    dftb_path = erepfit_input.options['toolchain']['path']

    for atom_symbol, atom_energy in erepfit_input.atomic_energy.items():
        if (type(atom_energy) != float):
            res = evaluate_atom(dftb_path, atom_energy, erepfit_input.electronic_slater_koster_files)
            erepfit_input.atomic_energy[atom_symbol] = res

    # size = len(erepfit_input.systems)
    for _, system in enumerate(tqdm(erepfit_input.systems)):
        _, elec_energy, elec_forces = evaluate_system(dftb_path, system, erepfit_input.electronic_slater_koster_files, erepfit_input.dftb_options)
        system.elec_data = {}
        system.elec_data['elec_energy'] = elec_energy
        system.elec_data['elec_forces'] = elec_forces
    

def test_dftb_binary(dftb_path):
    with tempfile.TemporaryDirectory() as tmp:
        res = subprocess.run([dftb_path], stdout=subprocess.PIPE, cwd=tmp)
        if (res.returncode == 0):
            for line in res.stdout.decode().splitlines():
                if ('DFTB+ paramerization version by Chien-Pin Chou:' in line):
                    version = int(line.split(':')[1])
                    if (version >= 20180919):
                        return True
        else:
            return False
    return False


from . import rep_spline
from . import rep_basis

from .rep_model import EnergyEquationDescription

import numpy as np
from sklearn import linear_model


class GeneralSolver():
    def __init__(self, solver=linear_model.LinearRegression()):
        self.solver = solver
    
    def solve(self, model, builder):
        #avaliable only for Spline model
        if (not isinstance(model.rep_models[0], rep_basis.BasisRepulsiveModel)):
            raise RuntimeError('The repulsive potential is not represented by basis functions')
        
        eqmat = np.concatenate([ x.equ_block for x in builder.equation_descriptions ])
        vref = np.concatenate([ x.ref_block for x in builder.equation_descriptions ])

        all_missing_atoms = list(set(sum([ list(x.missing_atom_counts.keys()) for x in builder.equation_descriptions if isinstance(x, EnergyEquationDescription) ], [])))
        final_mat = builder._extend_atomic_energy_equations(all_missing_atoms, eqmat)

        self.solver.fit(final_mat, vref)
        # #transform back for all coefficients
        a4 = self.solver.coef_[:eqmat.shape[1]]
        builder.fitted_atom_energy = { k: self.solver.coef_[eqmat.shape[1]+i] for i, k in enumerate(all_missing_atoms) }      

        model.resulting_repulsives = []

        shift = 0
        for ind in range(len(model.rep_names)):
            rep_name = model.rep_names[ind]
            rep_model = model.rep_models[ind]
            nvar = rep_model.num_variables

            coeffs = a4[shift:shift+nvar] 
            shift += nvar

            newrep = rep_basis.PolynomialRepulsive(rep_name, rep_model.cutoff, coeffs, rep_model.minimal_order, rep_model.start)
            spl_rep = newrep.to_spline4()
            model.resulting_repulsives.append(spl_rep)
        
        return a4

class ContinuitySolver():
    def __init__(self, continuous_order=3, solver=linear_model.LinearRegression()):
        self.continuous_order = continuous_order
        self.solver = solver

    def make_Q_R_S_matrix(self, dx, size):
        """ Get the continuity condition equations
        """
        mat_full = np.zeros( (size, size+1))
        mat_shift_full = np.zeros( (size, size+1))

        for i in range(0, size):
            for j in range(i, size+1):
                mat_shift_full[i, j] = 1.0*dx**(j-i)
                if (i==0):
                    mat_full[i, j] = 1.0
                else:
                    mat_full[i, j] = mat_full[i-1, j-1]*j

        final_mat = mat_full*mat_shift_full
        Q_mat = final_mat[0:size, 0:size]
        S_mat = final_mat[0:size, size-5:size+1]
        R_mat = np.diagflat(np.diag(Q_mat))*-1
        
        return Q_mat, R_mat, S_mat
    
    def make_Q_R_matrix(self, dx, nvar, order):
        nrows = order+1
        ncols = nvar
        mat_full = np.zeros( (nrows, ncols))
        mat_shift_full = np.zeros( (nrows, ncols))
        for i in range(0, nrows):
            for j in range(i, ncols):
                mat_shift_full[i, j] = 1.0*dx**(j-i)
                if (i==0):
                    mat_full[i, j] = 1.0
                else:
                    mat_full[i, j] = mat_full[i-1, j-1]*j

        final_mat = mat_full*mat_shift_full
        R_mat = np.diagflat(np.diag(final_mat))*-1
        
        return final_mat, R_mat

    def solve(self, model, builder):
        #avaliable only for Spline model
        if (not isinstance(model.rep_models[0], rep_spline.Spline4RepulsiveModel)):
            raise RuntimeError('The repulsive potential is not represented by Spline functions')

        #transform the equations using continuous conditions
        #define the dimension of the matrices
        nsplines = sum([ len(x.spline_models) for x in model.rep_models])
        nvar = model.rep_models[0].spline_models[0].num_variables
        block_size = self.continuous_order+1
        w_block_size = nvar-block_size
        mat_T = np.zeros( ((block_size*nsplines), (block_size*nsplines)))
        mat_W = np.zeros( ((block_size*nsplines), (w_block_size*nsplines)))

        ind = 0
        for spline_model in model.rep_models:
            knots = spline_model.knots
            for i in range(len(knots)-1):
                dx = knots[i+1] - knots[i]
                Q_mat, R_mat, S_mat = self.make_Q_R_S_matrix(dx, block_size)
                mat_T[ind*block_size:(ind+1)*block_size, ind*block_size:(ind+1)*block_size] = Q_mat
                if (i < len(knots)-2):
                    mat_T[ind*block_size:(ind+1)*block_size, (ind+1)*block_size:(ind+2)*block_size] = R_mat
                mat_W[ind*block_size:(ind+1)*block_size, (ind)*w_block_size:(ind+1)*w_block_size] = S_mat
                ind += 1

        eqmat = np.concatenate([ x.equ_block for x in builder.equation_descriptions ])
        vref = np.concatenate([ x.ref_block for x in builder.equation_descriptions ])
        
        #make eqmat separately for e3 and e4
        e3_index = sum([list(range(i, i+block_size )) for i in range(0, eqmat.shape[1]-block_size, nvar)], [])
        e3 = eqmat[:, e3_index]
        e4_index = sum([list(range(i, i+w_block_size )) for i in range(block_size, eqmat.shape[1], nvar)], [])
        e4 = eqmat[:, e4_index]       
        mat_T_inv_W = np.dot(np.linalg.inv(mat_T), -mat_W)
        mat_A  = np.dot(e3, mat_T_inv_W) + e4


        #adding missing atoms
        #get missing atom vector
        all_missing_atoms = list(set(sum([ list(x.missing_atom_counts.keys()) for x in builder.equation_descriptions if isinstance(x, EnergyEquationDescription) ], [])))
        final_mat = builder._extend_atomic_energy_equations(all_missing_atoms, mat_A)

        #solve with regularization solver
        self.solver.fit(final_mat, vref)

        #transform back for all coefficients

        a4 = self.solver.coef_[:mat_A.shape[1]]
        builder.fitted_atom_energy = { k: self.solver.coef_[mat_A.shape[1]+i] for i, k in enumerate(all_missing_atoms) }      
        a3 = np.dot(mat_T_inv_W, a4)


        model.resulting_repulsives = []
        vunknowns = []
        shift = 0
        for ind in range(len(model.rep_names)):
            rep_name = model.rep_names[ind]
            rep_model = model.rep_models[ind]
            knots = rep_model.knots

            coeffs = []
            for i in range(len(knots)-1):
                coeffs.append( np.append(a3[shift*block_size:shift*block_size+block_size], a4[shift:shift+w_block_size] ))
                vunknowns += a3[shift*block_size:shift*block_size+block_size].tolist()
                vunknowns += a4[shift:shift+w_block_size].tolist()
                shift += 1
            
            
            spl_rep = rep_spline.RepulsivePotenial(rep_name, knots, coeffs)
            model.resulting_repulsives.append(spl_rep)

        return np.array(vunknowns)

class ContinuityAdapter():
    def __init__(self, continuous_order=3):
        self.continuous_order = continuous_order
        

    def make_Q_R_S_matrix(self, dx, size):
        """ Get the continuity condition equations
        """
        mat_full = np.zeros( (size, size+1))
        mat_shift_full = np.zeros( (size, size+1))

        for i in range(0, size):
            for j in range(i, size+1):
                mat_shift_full[i, j] = 1.0*dx**(j-i)
                if (i==0):
                    mat_full[i, j] = 1.0
                else:
                    mat_full[i, j] = mat_full[i-1, j-1]*j

        final_mat = mat_full*mat_shift_full
        Q_mat = final_mat[0:size, 0:size]
        S_mat = final_mat[0:size, size-5:size+1]
        R_mat = np.diagflat(np.diag(Q_mat))*-1
        
        return Q_mat, R_mat, S_mat
    
    def make_Q_R_matrix(self, dx, nvar, order):
        nrows = order+1
        ncols = nvar
        mat_full = np.zeros( (nrows, ncols))
        mat_shift_full = np.zeros( (nrows, ncols))
        for i in range(0, nrows):
            for j in range(i, ncols):
                mat_shift_full[i, j] = 1.0*dx**(j-i)
                if (i==0):
                    mat_full[i, j] = 1.0
                else:
                    mat_full[i, j] = mat_full[i-1, j-1]*j

        final_mat = mat_full*mat_shift_full
        R_mat = np.diagflat(np.diag(final_mat))*-1
        
        return final_mat, R_mat

    def setup(self, model):
        #avaliable only for Spline model
        if (not isinstance(model.rep_models[0], rep_spline.Spline4RepulsiveModel)):
            raise RuntimeError('The repulsive potential is not represented by Spline functions')

        #transform the equations using continuous conditions
        #define the dimension of the matrices
        nsplines = sum([ len(x.spline_models) for x in model.rep_models])
        nvar = model.rep_models[0].spline_models[0].num_variables
        self.block_size = self.continuous_order+1
        self.w_block_size = nvar-self.block_size
        mat_T = np.zeros( ((self.block_size*nsplines), (self.block_size*nsplines)))
        mat_W = np.zeros( ((self.block_size*nsplines), (self.w_block_size*nsplines)))

        ind = 0
        for spline_model in model.rep_models:
            knots = spline_model.knots
            for i in range(len(knots)-1):
                dx = knots[i+1] - knots[i]
                Q_mat, R_mat, S_mat = self.make_Q_R_S_matrix(dx, self.block_size)
                mat_T[ind*self.block_size:(ind+1)*self.block_size, ind*self.block_size:(ind+1)*self.block_size] = Q_mat
                if (i < len(knots)-2):
                    mat_T[ind*self.block_size:(ind+1)*self.block_size, (ind+1)*self.block_size:(ind+2)*self.block_size] = R_mat
                mat_W[ind*self.block_size:(ind+1)*self.block_size, (ind)*self.w_block_size:(ind+1)*self.w_block_size] = S_mat
                ind += 1

        self.mat_T_inv_W = np.dot(np.linalg.inv(mat_T), -mat_W)
        self.model = model


    def transform(self, a4):
        a3 = np.dot(self.mat_T_inv_W, a4)

        vunknowns = []
        shift = 0
        for ind in range(len(self.model.rep_names)):
            rep_model = self.model.rep_models[ind]
            knots = rep_model.knots

            coeffs = []
            for i in range(len(knots)-1):
                coeffs.append( np.append(a3[shift*self.block_size:shift*self.block_size+self.block_size], a4[shift:shift+self.w_block_size] ))
                vunknowns += a3[shift*self.block_size:shift*self.block_size+self.block_size].tolist()
                vunknowns += a4[shift:shift+self.w_block_size].tolist()
                shift += 1
        return np.array(vunknowns)



# def solve_with_continuous_condition2(self, modelcollection, continuous_order=3):
        
    #     #avaliable only for Spline model
    #     if (not isinstance(modelcollection.rep_models[0], rep_spline.Spline4RepulsiveModel)):
    #         raise RuntimeError('The repulsive potential is not represented by Spline functions')
        
    #     #define the dimension of the matrices
    #     nsplines = sum([ len(x.spline_models) for x in modelcollection.rep_models])
    #     nvar = modelcollection.rep_models[0].spline_models[0].num_variables
        
    #     block_size = continuous_order+1
    #     w_block_size = nvar-block_size
    #     mat_C = np.zeros( ((block_size*nsplines), modelcollection.get_total_num_variables()) )
    #     vec_D = np.zeros( (block_size*nsplines) )               
        
    #     # print('Size of T and W ', mat_T.shape, mat_W.shape)
    #     # ref_vref = np.genfromtxt('../ref/vref.dat')
    #     # ref_eqmat = np.genfromtxt('../ref/eqmat.dat')

    #     ind = 0
    #     for spline_model in modelcollection.rep_models:
    #         knots = spline_model.knots
    #         for i in range(len(knots)-1):
    #             dx = knots[i+1] - knots[i]
    #             Q_mat, R_mat = self.make_Q_R_matrix(dx, nvar, continuous_order)
    #             mat_C[ind*block_size:(ind+1)*block_size, ind*nvar:(ind+1)*nvar] = Q_mat
                
    #             if (i < len(knots)-2):
    #                 mat_C[ind*block_size:(ind+1)*block_size, (ind+1)*nvar:(ind+1)*nvar+block_size] = R_mat

    #             ind += 1

    #     eqmat = np.concatenate([ x.equ_block for x in self.equation_descriptions ])
    #     vref = np.concatenate([ x.ref_block for x in self.equation_descriptions ])
        
    #     t, r, res, x, info = scipy.linalg.lapack.dgglse(eqmat, mat_C, vref, vec_D)

    #     print(x)
    #     modelcollection.resulting_repulsives = []
    #     shift = 0
    #     for ind in range(len(modelcollection.rep_names)):
    #         rep_name = modelcollection.rep_names[ind]
    #         rep_model = modelcollection.rep_models[ind]
    #         knots = rep_model.knots

    #         coeffs = []
    #         for i in range(len(knots)-1):
    #             coeffs.append( x[shift:shift+nvar] )
    #             shift += nvar
            
    #         spl_rep = spline_rep.RepulsivePotenial(rep_name, knots, coeffs)
    #         with open('{}.skf'.format(rep_name), 'w') as fout:
    #             print(spl_rep, file=fout)
    #         modelcollection.resulting_repulsives.append(spl_rep)


#!/usr/bin/env python3

import pymatgen as mg
import os
import uuid

import io

class SKFileInfo:
    def __init__(self, data):
        self.type2filenames = data.get('type2filenames', False)
        if (self.type2filenames):
            self.suffix = data.get('suffix', ".skf")
            self.prefix = data.get('prefix', "./")
            self.lowercase = data.get('lowercase', False)
            self.separator = data.get('separator', '-')
            self.max_angular_momentum = data.get('max_angular_momentum', dict())
        else:
            self.skfiles = data.get('skfiles', dict())

    
    def bool2YesNo(self, val):
        if (val):
            return "Yes"
        else:
            return "No"
  
    def to_hsd(self, fout):
        print('\n+Hamiltonian = +DFTB{', file=fout)        
        if (self.type2filenames):
            print('    !SlaterKosterFiles = !Type2Filenames{', file=fout)
            print('      !Prefix = "{}"'.format(self.prefix), file=fout)
            print('      !Suffix = "{}"'.format(self.suffix), file=fout)
            print('      !Separator = "{}"'.format(self.separator), file=fout)
            print('      !LowerCaseTypeName = {}'.format(self.bool2YesNo(self.lowercase)), file=fout)
            print('    }', file=fout)

            print('    !MaxAngularMomentum = {', file=fout)
            for k, v in self.max_angular_momentum.items():
                print('      {} = "{}"'.format(k, v), file=fout)
            print('    }', file=fout)

        else:
            print('    !SlaterKosterFiles = {', file=fout)
            for k, v in self.skfiles.items():
                print('      {} = {}'.format(k, v), file=fout)
            print('    }', file=fout)

            print('    !MaxAngularMomentum = {', file=fout)
            for k, v in self.max_angular_momentum.items():
                print('      {} = "{}"'.format(k, v), file=fout)
            print('    }', file=fout)
        print('}', file=fout)
    
    def adjust_path(self, basepath=os.curdir):
        if (self.type2filenames):
            if not os.path.isabs(self.prefix):
                self.prefix = os.path.join(basepath, self.prefix)
        else:
            l_maps = {}
            for k, v in self.skfiles.items():
                if os.path.isabs(v):
                    l_maps[k] = v
                else:
                    l_v = os.path.join(basepath, v)
                    l_maps[k] = l_v
            self.skfiles = l_maps

class Geometry:
    def __init__(self, jsondata):
        self.charge = jsondata.get("charge", 0)
        self.coordinates = jsondata.get("coordinates", list())

        if ('spin' in jsondata):
            self.spin = jsondata.get('spin', 1)
        if ('spinpol' in jsondata):
            self.spinpol = jsondata.get('spinpol', False)

        if ('lattice_vectors' in jsondata):
            self.lattice_vectors = jsondata['lattice_vectors']
            self.fractional_coordinates = jsondata['fractional_coordinates']
            self.scaling_factor = jsondata['scaling_factor']
            self.kpoints = jsondata['kpoints']

    def ispbc(self):
        return hasattr(self, 'lattice_vectors')

    def get_coordinate_struct(self):
        symbols = [x[0] for x in self.coordinates]
        coords = [x[1:4] for x in self.coordinates]
        if (hasattr(self, 'lattice_vectors')):
            lattice = mg.Lattice(self.lattice_vectors)
            mole = mg.Structure(lattice=lattice, species=symbols, coords=coords, coords_are_cartesian=(not self.fractional_coordinates))
        else:
            if (hasattr(self, 'spin')):
                mole = mg.Molecule(species=symbols, coords=coords, charge=self.charge, spin_multiplicity=self.spin)
            else:
                mole = mg.Molecule(species=symbols, coords=coords, charge=self.charge)
        return mole

class System:
    def __init__(self, jsondata):
        self.name = jsondata.get("name", "")
        self.template_input = jsondata.get("template_input", '')
        self.uuid = jsondata.get('uuid', uuid.uuid4() )
        self.geometry = Geometry(jsondata.get("geometry", dict()))

        if ('elec_data' in jsondata):
            self.elec_data =  jsondata['elec_data']

    def to_hsd(self, fout):
        # fout = io.StringIO()
        
        #write geometry
        symbols = [x[0] for x in self.geometry.coordinates]
        coords = [x[1:4] for x in self.geometry.coordinates]

        elements = list(set(symbols))
        pbc = False

        if (hasattr(self.geometry, 'lattice_vectors')):
            pbc = True

        print('!Geometry = !GenFormat{', file=fout)
        # fractional = False
        if (not pbc):
            print('{} C'.format(len(symbols)), file=fout)
        else:
            if (self.geometry.fractional_coordinates):
                print('{} F'.format(len(symbols)), file=fout)
                # fractional = True
            else:
                print('{} S'.format(len(symbols)), file=fout)
        print(' '.join(elements),file=fout)
        for i in range(len(symbols)):
            print('{} {} {:20.12F} {:20.12F} {:20.12F}'.format(
                i+1, elements.index(symbols[i])+1, *coords[i]
            ), file=fout)
        if (pbc):
            print("0.0 0.0 0.0", file=fout)
            for i in range(3):
                print('{:20.12F} {:20.12F} {:20.12F}'.format(
                *self.geometry.lattice_vectors[i]
            ), file=fout)
        print('}', file=fout)

        #write Hamiltonian
        print('\n+Hamiltonian = +DFTB{', file=fout)
        print('   !Charge = {}'.format(self.geometry.charge), file=fout)
        if (pbc):
            if ('supercell_folding' in self.geometry.kpoints):
                print('   !KPointsAndWeights = SupercellFolding{', file=fout)
                for line in self.geometry.kpoints['supercell_folding']:
                    print('    {} {} {}'.format(*line), file=fout)
                print('    }', file=fout)
            else:
                raise RuntimeError('Not implemented')
        print('}', file=fout)
            
class EnergyEquation:
    def __init__(self, jsondata):
        self.energy = jsondata['energy']
        self.unit = jsondata.get('unit', 'kcal/mol')
        self.name = jsondata['name']
        self.uuid = jsondata['uuid']
        self.weight = jsondata.get('weight', 1.0)

class ReactionItem:
    def __init__(self, jsondata):
        self.coefficient = jsondata['coefficient']
        self.name = jsondata['name']
        self.uuid = jsondata['uuid']
    
class ReactionEquation:
    def __init__(self, jsondata):
        self.energy = jsondata['energy']
        self.unit = jsondata.get('unit', 'kcal/mol')
        self.name = jsondata['name']
        self.uuid = jsondata['uuid']
        self.weight = jsondata.get('weight', 1.0)
        self.products = [ReactionItem(x) for x in jsondata['products']]
        self.reactants = [ReactionItem(x) for x in jsondata['reactants']]
    
    def __str__(self):
        fout = io.StringIO()
        print('Reaction: {}'.format(self.name), file=fout)
        print('Energy: {:20.12f} {:8s}'.format(self.energy, self.unit), file=fout)       
        print('Reactants and coefficients:', file=fout)
        for item in self.reactants:
            print('    {:8.3f} {}'.format(item.coefficient, item.name), file=fout)
        print('Products and coefficients:', file=fout)
        for item in self.products:
            print('    {:8.3f} {}'.format(item.coefficient, item.name), file=fout)
        return fout.getvalue()

    

class ForceEquation:
    def __init__(self, jsondata):       
        self.name = jsondata['name']
        self.uuid = jsondata['uuid']
        self.weight = jsondata.get('weight', 1.0)
        if ('reference_force') in jsondata:
            self.reference_force = jsondata['reference_force']
        if 'fitting_indexes' in jsondata:
            self.fitting_indexes = jsondata['fitting_indexes']

class AdditionalEquation:
    def __init__(self, jsondata):
        self.unit = jsondata.get('unit', 'bohr')
        self.distance = jsondata['distance']
        self.derivative = jsondata['derivative']
        self.value = jsondata['value']
        self.weight = jsondata.get('weight', 1.0)
        
class ErepfitInput:
    def __init__(self, jsondata):
        self.electronic_slater_koster_files = SKFileInfo(jsondata['electronic_slater_koster_files'])
        self.atomic_energy = jsondata.get('atomic_energy', dict())
        
        self.external_repulsive_potentials = jsondata.get('external_repulsive_potentials', dict())
        self.dftb_options = jsondata.get('dftb_options', dict())
        self.potential_grids = jsondata.get('potential_grids', dict())
        self.options = jsondata.get('options', dict())
        self.systems = [System(system) for system in jsondata.get('systems', list())]
        self.equations = {}
        node = jsondata.get('equations', dict())
        keywords = [
            ('energy', EnergyEquation),
            ('force', ForceEquation), 
            ('reaction', ReactionEquation)
        ]
        for key, clas in keywords:
            if (key in node):
                self.equations[key] = [clas(x) for x in node[key] ]
        if ('additional' in node):
             self.equations['additional'] = {'{}-{}'.format(*sorted(k.split('-'))): [AdditionalEquation(x) for x in v] for k, v in node['additional'].items() }




#!/usr/bin/env python3

from setuptools import setup, find_packages
import pyerepfit

with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')

setup(name='pyerepfit',
    version=pyerepfit.__version__,
    description='The python verion of the DFTB repulsive potential fitting tool, erepfit.',
    url='https://bitbucket.org/solccp/pyerepfit.git',
    author='Chien-Pin Chou',
    author_email='sol.chou@gmail.com',
    license='GPLv3',
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Chemistry",

    ],
    scripts=['bin/erepfit'],
    package_data={'pyerepfit': '*.json'},
    install_requires=install_requires,
    zip_safe=False)



#!/usr/bin/env python3

import os
import pymatgen as mg 
import pymatgen.core.units 

def which(file: str) -> str:
    for path in os.environ["PATH"].split(os.pathsep):
        if os.path.exists(os.path.join(path, file)):
            return os.path.join(path, file)

    return file

def get_energy_from_au(value: float, unit: str) -> float:
    if (unit.lower() == 'h'):
        return value
    elif (unit.lower() == 'ev'):
        return mg.core.units.Energy(value, 'Ha').to('eV')
    elif (unit.lower() == 'kcal/mol'):
        return value*627.50947415
    else:
        raise ValueError('Unknown unit: {}'.format(unit))


def get_energy_in_au(value: float, unit: str) -> float:
    if (unit.lower() == 'h'):
        return value
    elif (unit.lower() == 'ev'):
        return mg.core.units.Energy(value, 'eV').to('Ha')
    elif (unit.lower() == 'kcal/mol'):
        return value/627.50947415
    else:
        raise ValueError('Unknown unit: {}'.format(unit))

def get_length_in_au(value: float, unit: str) -> float:
    if (unit.lower() == 'bohr'):
        return value
    elif (unit.lower() == 'aa'):
        return value*mg.core.units.ang_to_bohr
    else:
        raise ValueError('Unknown unit: {}'.format(unit))
